import typing
from .schema import Compiler


M = typing.TypeVar('M')


@typing.overload
def modelclass(
    cls: typing.Type[M],
    *, compiler: Compiler = None,
    immutable: bool = False,
) -> typing.Type[M]:
    ...


@typing.overload
def modelclass(
    *, compiler: Compiler = None,
    immutable: bool = False,
) -> typing.Callable[[typing.Type[M]], typing.Type[M]]:
    ...


def fields(m: typing.Any) -> typing.Set[str]:
    ...


def asdict(
    m: typing.Any,
    *, keys: typing.Iterable[str] = None,
) -> typing.Dict[str, typing.Any]:
    ...



import re
import sys
import uuid
import time
import datetime
import ipaddress
import typing
from copy import copy
from functools import partial
from urllib.parse import urlparse, urlunparse

from .exception import Invalid, SchemaError, mark_key, mark_index
from ._vendor import durationpy
from ._vendor.email_validator import validate_email, EmailNotValidError
from ._vendor.fqdn import FQDN


def is_dict(obj):
    # use isinstance(obj, Mapping) is slow,
    # hasattr check can speed up about 30%
    return hasattr(obj, '__getitem__') and hasattr(obj, 'get')


def get_dict_value(obj, key):
    return obj.get(key, None)


def get_object_value(obj, key):
    return getattr(obj, key, None)


def _is_empty(value):
    return value is None or value == ''


def _update_validate_func_info(validate_func, origin_func, schema):
    # make friendly validate func representation
    m_repr = schema.repr(prefix=False, desc=False)
    validate_func.__schema__ = schema
    validate_func.__module__ = origin_func.__module__
    validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
    if hasattr(origin_func, '__qualname__'):
        qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
        validate_func.__qualname__ = qualname
    if origin_func.__doc__:
        validate_func.__doc__ = origin_func.__doc__
    else:
        validate_func.__doc__ = schema.params.get('desc')


def _parse_hints(hints):
    if not isinstance(hints, (tuple, set, list)):
        hints = [hints]
    is_string = False
    is_object = False
    types = []
    for hint in hints:
        if hint is str:
            is_string = True
        else:
            is_object = True
        if hint is object:
            hint = typing.Any
        types.append(hint)
    return is_string, is_object, tuple(types)


def _update_validate_func_type_hints(
        validate_func,
        optional, has_default,
        accept_hints, output_hints,
):
    assert accept_hints, 'no accept_hints'
    assert output_hints, 'no output_hints'
    if len(accept_hints) == 1:
        value_typing = accept_hints[0]
    else:
        value_typing = typing.Union[accept_hints]
    if optional:
        value_typing = typing.Optional[value_typing]
    if len(output_hints) == 1:
        return_typing = output_hints[0]
    else:
        return_typing = typing.Union[output_hints]
    if optional and not has_default:
        return_typing = typing.Optional[return_typing]
    annotations = {'value': value_typing, 'return': return_typing}
    validate_func.__annotations__ = annotations


def validator(string=None, *, accept=None, output=None):
    """Decorator for create validator

    It will handle params default,optional,desc automatically.

    Usage:

        @validator(accept=(str,object), output=(str, object))
        def xxx_validator(compiler, output_object, **params):
            def validate(value):
                try:
                    # validate/convert the value
                except Exception:
                    # raise Invalid('invalid xxx')
                if output_object:
                    # return python object
                else:
                    # return string
            return validate

    Args:
        accept (str | object | (str,object)):
            str: the validator accept only string, treat both None and empty string as None
            object: the validator accept only object
            (str,object): (default) the validator accept both string and object,
                treat both None and empty string as None
        output (str | object | (str,object)):
            str: (default) the validator always output string, convert None to empty string
            object: the validator always output object
            (str, object): the validator can output both string and object,
                and has an `object` parameter to control which to output
        string (bool): deprecated in v1.1.0.
            string=True equal to accept=(str, object), output=str
            string=False equal to accept=(str, object), output=object
    """
    # cdef bint accept_string, accept_object, output_string, output_object
    if accept is None or not accept:
        accept_string = True
        accept_object = True
        accept_hints = (str, typing.Any)
    else:
        accept_string, accept_object, accept_hints = _parse_hints(accept)
    if not (accept_string or accept_object):
        raise ValueError('invalid accept argument {}'.format(accept))
    if output is None or not output:
        output_string = string
        output_object = not string
        output_hints = (str,) if string else (typing.Any,)
    else:
        output_string, output_object, output_hints = _parse_hints(output)
    if not (output_string or output_object):
        raise ValueError('invalid output argument {}'.format(output))
    del string, accept, output

    def decorator(f):

        def _m_validator(compiler, schema):
            params = schema.params.copy()
            if schema.items is not None:
                params['items'] = schema.items
            local_output_object = output_object
            if output_string and output_object:
                local_output_object = bool(params.get('object', None))
                # TODO: fix special case of timedelta
                if schema.validator == 'timedelta':
                    local_output_object = not bool(params.get('string', None))
            if output_object and 'object' in params:
                params['output_object'] = bool(params.pop('object', None))
            if local_output_object:
                null_output = None
            else:
                null_output = ''
            optional = params.pop('optional', False)
            default = params.pop('default', None)
            params.pop('desc', None)
            invalid_to_default = params.pop('invalid_to_default', False)
            has_invalid_to = 'invalid_to' in params
            invalid_to = params.pop('invalid_to', None)
            # cdef bint has_default
            if accept_string:
                has_default = not (default is None or default == '')
            else:
                has_default = not (default is None)
            if has_invalid_to and invalid_to_default:
                raise SchemaError('can not set both invalid_to and invalid_to_default')
            if invalid_to_default and (not has_default) and (not optional):
                raise SchemaError('default or optional must be set when set invalid_to_default')
            try:
                validate = f(compiler, **params)
            except TypeError as e:
                raise SchemaError(str(e)) from None
            # check default value
            if has_default:
                try:
                    default = validate(default)
                except Invalid:
                    msg = 'invalid default value {!r}'.format(default)
                    raise SchemaError(msg) from None
                if invalid_to_default:
                    invalid_to = default
            else:
                if invalid_to_default:
                    invalid_to = null_output
            # check invalid_to value
            if has_invalid_to:
                try:
                    invalid_to = validate(invalid_to)
                except Invalid:
                    msg = 'invalid invalid_to value {!r}'.format(invalid_to)
                    raise SchemaError(msg) from None

            # check null, empty string and default
            def _m_validate(value):
                # cdef bint is_null
                if accept_string:
                    is_null = _is_empty(value)
                else:
                    is_null = value is None
                if is_null:
                    if has_default:
                        return default
                    elif optional:
                        return null_output
                    else:
                        raise Invalid('required')
                if not accept_object and not isinstance(value, str):
                    raise Invalid('require string value')
                value = validate(value)
                # check again after validate
                if accept_string:
                    is_null = _is_empty(value)
                else:
                    is_null = value is None
                if is_null:
                    if has_default:
                        return default
                    elif optional:
                        return null_output
                    else:
                        raise Invalid('required')
                return value

            supress_invalid = has_invalid_to or invalid_to_default

            def m_validate(value):
                try:
                    return _m_validate(value)
                except Invalid as ex:
                    ex.set_value(value)
                    if supress_invalid:
                        return invalid_to
                    else:
                        raise

            _update_validate_func_info(m_validate, f, schema)

            _update_validate_func_type_hints(
                m_validate, optional=optional, has_default=has_default,
                accept_hints=accept_hints, output_hints=output_hints)

            return m_validate

        def m_validator(compiler, schema):
            try:
                return _m_validator(compiler, schema)
            except SchemaError as ex:
                ex.set_value(schema)
                raise

        # TODO: deprecate below attributes because they are implement details
        m_validator.is_string = output_string
        m_validator.accept_string = accept_string
        m_validator.accept_object = accept_object
        m_validator.output_string = output_string
        m_validator.output_object = output_object
        m_validator.validator = f
        # end deprecate

        m_validator.__module__ = f.__module__
        m_validator.__name__ = f.__name__
        if hasattr(f, '__qualname__'):
            m_validator.__qualname__ = f.__qualname__
        m_validator.__doc__ = f.__doc__
        return m_validator
    return decorator


_UNIQUE_CHECK_ERROR_MESSAGE = "unable to check unique for non-hashable types"


def _key_of_scalar(v):
    return v


def _key_func_of_schema(schema):
    if schema is None:
        raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)

    if schema.validator == 'dict':
        if schema.items is None:
            raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
        keys = []
        for k, v in schema.items.items():
            keys.append((k, _key_func_of_schema(v)))

        def key_of(v):
            # cdef str k
            return tuple(key_of_value(v[k]) for k, key_of_value in keys)

    elif schema.validator == 'list':
        if schema.items is None:
            raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
        key_of_value = _key_func_of_schema(schema.items)

        def key_of(v):
            return tuple(key_of_value(x) for x in v)

    else:
        key_of = _key_of_scalar

    return key_of


@validator(accept=typing.Iterable, output=typing.List)
def list_validator(compiler, items=None, minlen=0, maxlen=1024,
                   unique=False):
    if items is None:
        inner = None
    else:
        with mark_index():
            inner = compiler.compile(items)
    if unique:
        key_of = _key_func_of_schema(items)
    del compiler, items

    def validate(value):
        try:
            value = enumerate(value)
        except TypeError:
            raise Invalid('not list')
        result = []
        if unique:
            keys = set()
        i = -1
        for i, x in value:
            if i >= maxlen:
                raise Invalid('list length must <= %d' % maxlen)
            with mark_index(i):
                v = inner(x) if inner is not None else copy(x)
                if unique:
                    k = key_of(v)
                    if k in keys:
                        raise Invalid('not unique')
                    keys.add(k)
            result.append(v)
        if minlen > 0 and i + 1 < minlen:
            raise Invalid('list length must >= %d' % minlen)
        return result
    return validate


def _slim_dict(value):
    return {k: v for k, v in value.items() if not _is_empty(v)}


@validator(accept=(typing.Mapping, typing.Any), output=dict)
def dict_validator(compiler, items=None, key=None, value=None,
                   minlen=0, maxlen=1024, slim=False):
    if items is None:
        inners = None
    else:
        inners = []
        for k, v in items.items():
            with mark_key(k):
                inners.append((k, compiler.compile(v)))
    validate_extra_key = validate_extra_value = None
    if key is not None:
        with mark_key('$self_key'):
            validate_extra_key = compiler.compile(key)
    if value is not None:
        with mark_key('$self_value'):
            validate_extra_value = compiler.compile(value)
    # cdef bint is_dynamic
    is_dynamic = bool(validate_extra_key or validate_extra_value)
    del compiler, items, key, value

    def validate(value):
        if inners is None and not is_dynamic:
            if not is_dict(value):
                raise Invalid('must be dict')
            if len(value) > maxlen:
                raise Invalid('dict length must <= %d' % maxlen)
            elif minlen > 0 and len(value) < minlen:
                raise Invalid('dict length must >= %d' % minlen)
            if slim:
                value = _slim_dict(value)
            return copy(value)
        if is_dict(value):
            getter = get_dict_value
            if is_dynamic:
                if len(value) > maxlen:
                    raise Invalid('dict length must <= %d' % maxlen)
                elif minlen > 0 and len(value) < minlen:
                    raise Invalid('dict length must >= %d' % minlen)
        else:
            getter = get_object_value
            if is_dynamic:
                raise Invalid("dynamic dict not allowed non-dict value")
        result = {}
        # cdef str k
        if inners is not None:
            for k, inner in inners:
                with mark_key(k):
                    result[k] = inner(getter(value, k))
        if is_dynamic:
            extra_keys = map(str, set(value) - set(result))
            for k in extra_keys:
                if validate_extra_key:
                    with mark_key('$self_key'):
                        k = str(validate_extra_key(k))
                with mark_key(k):
                    v = getter(value, k)
                    if validate_extra_value is not None:
                        result[k] = validate_extra_value(v)
                    else:
                        result[k] = copy(v)
        if slim:
            result = _slim_dict(result)
        return result
    return validate


@validator(accept=(typing.Mapping, typing.Any), output=object)
def model_validator(compiler, items=None):
    if items is None:
        raise SchemaError('model class not provided')

    def validate(value):
        return items(value)

    return validate


def _dump_enum_value(value):
    if value is None:
        return 'null'
    elif value is False:
        return 'false'
    elif value is True:
        return 'true'
    elif isinstance(value, str):
        return repr(value)  # single quotes by default
    else:
        return str(value)  # number


@validator(output=object)
def enum_validator(compiler, items):
    if not items:
        raise SchemaError("enum items not provided")
    expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'
    items = frozenset(items)

    def validate(value):
        if value in items:
            return value
        raise Invalid("expect one of {}".format(expects))

    return validate


def union_validator(compiler, schema):
    if not schema.items:
        raise SchemaError('union schemas not provided')
    default = schema.params.get('default')
    if default is not None:
        raise SchemaError("not allowed default for union schema")
    by = schema.params.get('by')
    if isinstance(schema.items, list):
        if by is not None:
            raise SchemaError("not allowed 'by' argument for union list schema")
        return _union_list_validator(compiler, schema)
    elif isinstance(schema.items, dict):
        if by is None or by == "":
            raise SchemaError("required 'by' argument for union dict schema")
        if not isinstance(by, str):
            raise SchemaError("'by' argument must be str type for union schema")
        return _union_dict_validator(compiler, schema)
    else:
        raise SchemaError('union schemas type invalid')


def _optional_or_has_default(schema):
    if schema.params.get('optional'):
        return True
    if schema.params.get('default') is not None:
        return True
    return False


def _union_list_validator(compiler, schema):
    scalar_inner = None
    list_inner = None
    dict_inner = None
    for i, inner_schema in enumerate(schema.items):
        with mark_index(i):
            if inner_schema.validator == 'union':
                raise SchemaError('ambiguous union schema')
            if _optional_or_has_default(inner_schema):
                raise SchemaError('not allowed optional or default for union schemas')
            if schema.params.get('optional'):
                inner_schema = inner_schema.copy()
                inner_schema.params['optional'] = True
            if inner_schema.validator == 'list':
                if list_inner is not None:
                    raise SchemaError('ambiguous union schema')
                list_inner = compiler.compile(inner_schema)
            elif inner_schema.validator in ('dict', 'model'):
                if dict_inner is not None:
                    raise SchemaError('ambiguous union schema')
                dict_inner = compiler.compile(inner_schema)
            else:
                if scalar_inner is not None:
                    raise SchemaError('ambiguous union schema')
                scalar_inner = compiler.compile(inner_schema)

    def validate(value):
        if isinstance(value, list):
            if list_inner is None:
                raise Invalid('not allowed list')
            return list_inner(value)
        elif is_dict(value) or hasattr(value, '__asdict__'):
            if dict_inner is None:
                raise Invalid('not allowed dict')
            return dict_inner(value)
        elif value is None:
            return (scalar_inner or list_inner or dict_inner)(value)
        else:
            if scalar_inner is None:
                raise Invalid('not allowed scalar value')
            return scalar_inner(value)

    _update_validate_func_info(validate, union_validator, schema)

    return validate


@validator(accept=object, output=object)
def _union_dict_validator(compiler, items, by):
    inners = {}
    for key, schema in items.items():
        with mark_key(key):
            if schema.validator not in ('dict', 'model'):
                raise SchemaError('must be dict or model schema')
            if _optional_or_has_default(schema):
                raise SchemaError('not allowed optional or default for union schemas')
            is_model = schema.validator == 'model'
            inners[key] = (is_model, compiler.compile(schema))
    expect_bys = '{' + ', '.join(sorted(inners.keys())) + '}'
    del key, schema, is_model

    def validate(value):
        if is_dict(value):
            getter = get_dict_value
        else:
            getter = get_object_value
        # cdef str by_name
        with mark_key(by):
            by_name = getter(value, by)
            if not by_name:
                raise Invalid('required', value=by_name)
            inner_info = inners.get(by_name)
            if inner_info is None:
                err_msg = 'expect one of {}'.format(expect_bys)
                raise Invalid(err_msg, value=by_name)
        is_model, inner = inner_info
        result = inner(value)
        if not is_model:
            result[by] = by_name
        return result

    return validate


def any_validate(value):
    return copy(value)


@validator(accept=object, output=object)
def any_validator(compiler, **ignore_kwargs):
    """Accept any value"""
    return any_validate


@validator(accept=(int, float, str), output=int)
def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):
    """Validate int or convert string to int

    Args:
        min (int): the min value, default -sys.maxsize
        max (int): the max value, default sys.maxsize
    """
    min, max = int(min), int(max)

    def validate(value):
        try:
            v = int(value)
        except Exception:
            raise Invalid('invalid int') from None
        if v < min:
            raise Invalid('value must >= %d' % min)
        elif v > max:
            raise Invalid('value must <= %d' % max)
        return v
    return validate


_TRUE_VALUES = {
    True, 1, '1',
    'True', 'true', 'TRUE',
    'Yes', 'yes', 'YES', 'y', 'Y',
    'On', 'on', 'ON',
}
_FALSE_VALUES = {
    False, 0, '0',
    'False', 'false', 'FALSE',
    'No', 'no', 'NO', 'n', 'N',
    'Off', 'off', 'OFF',
}


@validator(accept=(bool, int, str), output=bool)
def bool_validator(compiler):
    """Validate bool"""
    def validate(value):
        if value in _TRUE_VALUES:
            return True
        elif value in _FALSE_VALUES:
            return False
        else:
            raise Invalid('invalid bool')
    return validate


@validator(accept=(int, float, str), output=float)
def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,
                    exmin=False, exmax=False):
    """Validate float string

    Args:
        min (float): the min value, default -sys.float_info.max
        max (float): the max value, default sys.float_info.max
        exmin (bool,float): exclude min value or not, default false
        exmax (bool,float): exclude max value or not, default false
    """
    min, max = float(min), float(max)
    if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):
        min = float(exmin)
        exmin = True
    else:
        exmin = bool(exmin)
    if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):
        max = float(exmax)
        exmax = True
    else:
        exmax = bool(exmax)

    def validate(value):
        try:
            v = float(value)
        except Exception:
            raise Invalid('invalid float') from None
        if exmin:
            if v <= min:
                raise Invalid('value must > %d' % min)
        else:
            if v < min:
                raise Invalid('value must >= %d' % min)
        if exmax:
            if v >= max:
                raise Invalid('value must < %d' % max)
        else:
            if v > max:
                raise Invalid('value must <= %d' % max)
        return v
    return validate


@validator(accept=(str, object), output=str)
def str_validator(compiler, minlen=0, maxlen=1024 * 1024,
                  strip=False, escape=False, match=None,
                  accept_object=False):
    """Validate string

    Args:
        minlen (int): min length of string, default 0
        maxlen (int): max length of string, default 1024*1024
        strip (bool): strip white space or not, default false
        escape (bool): escape to html safe string or not, default false
        match (str): regex to match, default None
    """

    # To make sure that the entire string matches
    if match:
        try:
            re_match = re.compile(r'(?:%s)\Z' % match).match
        except Exception as ex:
            raise SchemaError('match regex %s compile failed' % match) from ex
    else:
        re_match = None

    def validate(value):
        if not isinstance(value, str):
            if accept_object or isinstance(value, int):
                value = str(value)
            else:
                raise Invalid('invalid string')
        if strip:
            value = value.strip()
        length = len(value)
        if length < minlen:
            raise Invalid('string length must >= %d' % minlen)
        elif length > maxlen:
            raise Invalid('string length must <= %d' % maxlen)
        if escape:
            value = (value.replace('&', '&amp;')
                     .replace('>', '&gt;')
                     .replace('<', '&lt;')
                     .replace("'", '&#39;')
                     .replace('"', '&#34;'))
        if re_match is not None and not re_match(value):
            raise Invalid("string not match regex %s" % match)
        return value
    return validate


@validator(accept=bytes, output=bytes)
def bytes_validator(compiler, minlen=0, maxlen=-1):
    """Validate bytes

    Args:
        minlen (int): min length of bytes, default 0
        maxlen (int): max length of bytes, default unlimited
    """

    def validate(value):
        if not isinstance(value, bytes):
            raise Invalid('invalid bytes')
        length = len(value)
        if length < minlen:
            raise Invalid('bytes length must >= %d' % minlen)
        elif maxlen > -1 and length > maxlen:
            raise Invalid('bytes length must <= %d' % maxlen)
        return value

    return validate


@validator(accept=(str, datetime.date), output=(str, datetime.date))
def date_validator(compiler, format='%Y-%m-%d', output_object=False):
    """Validate date string or convert date to string

    Args:
        format (str): date format, default ISO8601 format
    """
    def validate(value):
        try:
            if not isinstance(value, (datetime.datetime, datetime.date)):
                value = datetime.datetime.strptime(value, format)
            if isinstance(value, datetime.datetime):
                value = value.date()
            if output_object:
                return value
            else:
                return value.strftime(format)
        except Exception:
            raise Invalid('invalid date') from None
    return validate


@validator(accept=(str, datetime.time), output=(str, datetime.time))
def time_validator(compiler, format='%H:%M:%S', output_object=False):
    """Validate time string or convert time to string

    Args:
        format (str): time format, default ISO8601 format
    """
    def validate(value):
        try:
            if not isinstance(value, (datetime.datetime, datetime.time)):
                value = datetime.datetime.strptime(value, format)
            if isinstance(value, datetime.datetime):
                value = value.time()
            if output_object:
                return value
            else:
                return value.strftime(format)
        except Exception:
            raise Invalid('invalid time') from None
    return validate


@validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
def datetime_validator(compiler, format='%Y-%m-%dT%H:%M:%S.%fZ', output_object=False):
    """Validate datetime string or convert datetime to string

    Args:
        format (str): datetime format, default ISO8601 format
    """
    def validate(value):
        try:
            if isinstance(value, tuple):
                value = datetime.datetime.fromtimestamp(time.mktime(value))
            elif not isinstance(value, datetime.datetime):
                value = datetime.datetime.strptime(value, format)
            if output_object:
                return value
            else:
                return value.strftime(format)
        except Exception:
            raise Invalid('invalid datetime') from None
    return validate


def _parse_timedelta(value):
    if isinstance(value, (int, float)):
        value = datetime.timedelta(seconds=value)
    elif isinstance(value, str):
        value = durationpy.from_str(value)
    else:
        if not isinstance(value, datetime.timedelta):
            raise ValueError("invalid timedelta")
    return value


@validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
def timedelta_validator(compiler, min=None, max=None, string=False,
                        extended=False, output_object=False):
    """Validate timedelta string or convert timedelta to string

    Format (Go's Duration strings):
        ns - nanoseconds
        us - microseconds
        ms - milliseconds
        s - seconds
        m - minutes
        h - hours
        d - days
        mo - months
        y - years
    """
    if string and output_object:
        raise SchemaError('can not output both string and object')
    try:
        min_value = _parse_timedelta(min) if min is not None else None
    except (durationpy.DurationError, ValueError, TypeError) as ex:
        raise SchemaError('invalid min timedelta') from ex
    try:
        max_value = _parse_timedelta(max) if max is not None else None
    except (durationpy.DurationError, ValueError, TypeError) as ex:
        raise SchemaError('invalid max timedelta') from ex
    del min, max
    min_repr = max_repr = None
    if min_value is not None:
        min_repr = durationpy.to_str(min_value, extended=True)
    if max_value is not None:
        max_repr = durationpy.to_str(max_value, extended=True)

    def validate(value):
        try:
            value = _parse_timedelta(value)
        except (durationpy.DurationError, ValueError, TypeError) as ex:
            raise Invalid('invalid timedelta') from ex
        if min_value is not None:
            if value < min_value:
                raise Invalid('value must >= {}'.format(min_repr))
        if max_value is not None:
            if value > max_value:
                raise Invalid('value must <= {}'.format(max_repr))
        if output_object:
            return value
        if string:
            value = durationpy.to_str(value, extended=extended)
        else:
            value = value.total_seconds()
        return value
    return validate


@validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
def ipv4_validator(compiler, output_object=False):
    def validate(value):
        try:
            value = ipaddress.IPv4Address(value.strip())
        except ipaddress.AddressValueError as ex:
            raise Invalid(str(ex)) from None
        except Exception:
            raise Invalid('invalid ipv4 address') from None
        if output_object:
            return value
        else:
            return value.compressed
    return validate


@validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
def ipv6_validator(compiler, output_object=False):
    def validate(value):
        try:
            value = ipaddress.IPv6Address(value.strip())
        except ipaddress.AddressValueError as ex:
            raise Invalid(str(ex)) from None
        except Exception:
            raise Invalid('invalid ipv6 address') from None
        if output_object:
            return value
        else:
            return value.compressed
    return validate


@validator(accept=str, output=str)
def email_validator(compiler):
    # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
    # http://emailregex.com/
    # https://github.com/JoshData/python-email-validator
    _validate = partial(
        validate_email,
        allow_smtputf8=False,
        check_deliverability=False,
        allow_empty_local=False,
    )

    def validate(value):
        try:
            value = _validate(value.strip())
        except EmailNotValidError as ex:
            raise Invalid(str(ex)) from None
        except Exception:
            raise Invalid('invalid email address') from None
        return value['email']
    return validate


@validator(output=(str, object))
def url_validator(compiler, scheme='http https', maxlen=255, output_object=False):
    # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
    # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
    # https://github.com/python-hyper/rfc3986
    # https://github.com/dgerber/rfc3987
    # https://github.com/tkem/uritools
    allow_scheme = set(scheme.split())

    def validate(value):
        try:
            value = value.strip()
        except Exception:
            raise Invalid('invalid url') from None
        if len(value) > maxlen:
            raise Invalid('url length must <= {}'.format(maxlen))
        try:
            parsed = urlparse(value)
        except Exception:
            raise Invalid('invalid url') from None
        if not parsed.scheme or parsed.scheme not in allow_scheme:
            raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
        if output_object:
            return parsed
        else:
            return urlunparse(parsed)
    return validate


@validator(output=str)
def fqdn_validator(compiler):
    def validate(value):
        try:
            value = value.strip()
            fqdn_obj = FQDN(value)
            if fqdn_obj.is_valid:
                return fqdn_obj.relative
        except (ValueError, TypeError, AttributeError) as ex:
            raise Invalid("invalid fqdn") from ex
        raise Invalid("invalid fqdn")
    return validate


@validator(output=(str, uuid.UUID))
def uuid_validator(compiler, version=None, output_object=False):
    if version is None:
        msg = 'invalid uuid'
    else:
        if version not in {1, 3, 4, 5}:
            raise SchemaError('illegal version number')
        msg = 'invalid uuid{}'.format(version)

    def validate(value):
        if not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value.strip())
            except Exception:
                raise Invalid(msg) from None
        if version is not None and value.version != version:
            raise Invalid(msg)
        if output_object:
            return value
        else:
            return str(value)
    return validate


def create_re_validator(name, r, default_maxlen=255):
    """Create validator by regex string

    It will make sure that the entire string matches, so needn't
    add `^`,`$` to regex string.

    Args:
        name (str): validator name, used in error message
        r (str): regex string
    """
    # To make sure that the entire string matches
    match = re.compile(r'(?:%s)\Z' % r).match
    message = 'invalid %s' % name

    def re_validator(compiler, minlen=0, maxlen=default_maxlen, strip=False):
        def validate(value):
            if not isinstance(value, str):
                raise Invalid('value must be string')
            if strip:
                value = value.strip()
            length = len(value)
            if length < minlen:
                raise Invalid('%s length must >= %d' % (name, minlen))
            elif length > maxlen:
                raise Invalid('%s length must <= %d' % (name, maxlen))
            if match(value):
                return value
            else:
                raise Invalid(message)
        return validate
    re_validator.__name__ = name + '_validator'
    re_validator.__qualname__ = name + '_validator'
    return validator(accept=str, output=str)(re_validator)


builtin_validators = {
    'list': list_validator,
    'dict': dict_validator,
    'model': model_validator,
    'union': union_validator,
    'enum': enum_validator,
    'any': any_validator,
    'int': int_validator,
    'bool': bool_validator,
    'float': float_validator,
    'str': str_validator,
    'bytes': bytes_validator,
    'date': date_validator,
    'time': time_validator,
    'datetime': datetime_validator,
    'timedelta': timedelta_validator,
    'ipv4': ipv4_validator,
    'ipv6': ipv6_validator,
    'email': email_validator,
    'url': url_validator,
    'fqdn': fqdn_validator,
    'uuid': uuid_validator,
}

regexs = {
    'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),
    'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),
    'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
}
for name, options in regexs.items():
    builtin_validators[name] = create_re_validator(name, *options)


def create_enum_validator(name, items, string=True):
    """Create validator by enum items

    Args:
        name (str): validator name, used in error message
        items (iterable): enum items
        string (bool): is string like or not

    Deprecated since v1.2.0, use enum validator instead.
    """
    items = set(items)
    message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))

    def enum_validator(compiler):
        def validate(value):
            if value in items:
                return value
            raise Invalid(message)
        return validate
    enum_validator.__name__ = name + '_validator'
    enum_validator.__qualname__ = name + '_validator'
    if string:
        return validator(accept=str, output=str)(enum_validator)
    else:
        return validator(accept=object, output=object)(enum_validator)



"""
https://github.com/JoshData/python-email-validator/blob/master/email_validator/__init__.py

Changes: remove check_deliverability, remove allow_smtputf8
"""
import sys
import re
import idna  # implements IDNA 2008; Python's codec is only IDNA 2003


# Based on RFC 2822 section 3.2.4 / RFC 5322 section 3.2.3, these
# characters are permitted in email addresses (not taking into
# account internationalization):
ATEXT = r'a-zA-Z0-9_!#\$%&\'\*\+\-/=\?\^`\{\|\}~'

# A "dot atom text", per RFC 2822 3.2.4:
DOT_ATOM_TEXT = '[' + ATEXT + ']+(?:\\.[' + ATEXT + ']+)*'

# RFC 6531 section 3.3 extends the allowed characters in internationalized
# addresses to also include three specific ranges of UTF8 defined in
# RFC3629 section 4, which appear to be the Unicode code points from
# U+0080 to U+10FFFF.
ATEXT_UTF8 = ATEXT + u"\u0080-\U0010FFFF"
DOT_ATOM_TEXT_UTF8 = '[' + ATEXT_UTF8 + ']+(?:\\.[' + ATEXT_UTF8 + ']+)*'

# The domain part of the email address, after IDNA (ASCII) encoding,
# must also satisfy the requirements of RFC 952/RFC 1123 which restrict
# the allowed characters of hostnames further. The hyphen cannot be at
# the beginning or end of a *dot-atom component* of a hostname either.
ATEXT_HOSTNAME = r'(?:(?:[a-zA-Z0-9][a-zA-Z0-9\-]*)?[a-zA-Z0-9])'

# ease compatibility in type checking
if sys.version_info >= (3,):
    unicode_class = str
else:
    unicode_class = unicode  # noqa: F821

    # turn regexes to unicode (because 'ur' literals are not allowed in Py3)
    ATEXT = ATEXT.decode("ascii")
    DOT_ATOM_TEXT = DOT_ATOM_TEXT.decode("ascii")
    ATEXT_HOSTNAME = ATEXT_HOSTNAME.decode("ascii")

DEFAULT_TIMEOUT = 15  # secs


class EmailNotValidError(ValueError):
    """Parent class of all exceptions raised by this module."""
    pass


class EmailSyntaxError(EmailNotValidError):
    """Exception raised when an email address fails validation because of its form."""
    pass


class EmailUndeliverableError(EmailNotValidError):
    """Exception raised when an email address fails validation because its domain name does not appear deliverable."""
    pass


def validate_email(
    email,
    allow_smtputf8=True,
    allow_empty_local=False,
    check_deliverability=True,
    timeout=DEFAULT_TIMEOUT,
):
    """
    Validates an email address, raising an EmailNotValidError if the address is not valid or returning a dict of
    information when the address is valid. The email argument can be a str or a bytes instance,
    but if bytes it must be ASCII-only.
    """

    # Allow email to be a str or bytes instance. If bytes,
    # it must be ASCII because that's how the bytes work
    # on the wire with SMTP.
    if not isinstance(email, (str, unicode_class)):
        try:
            email = email.decode("ascii")
        except ValueError:
            raise EmailSyntaxError("The email address is not valid ASCII.")

    # At-sign.
    parts = email.split('@')
    if len(parts) != 2:
        raise EmailSyntaxError("The email address is not valid. It must have exactly one @-sign.")

    # Prepare a dict to return on success.
    ret = {}

    # Validate the email address's local part syntax and update the return
    # dict with metadata.
    ret.update(validate_email_local_part(parts[0], allow_smtputf8=allow_smtputf8, allow_empty_local=allow_empty_local))

    # Validate the email address's domain part syntax and update the return
    # dict with metadata.
    ret.update(validate_email_domain_part(parts[1]))

    # If the email address has an ASCII form, add it.
    ret["email"] = ret["local"] + "@" + ret["domain_i18n"]
    if not ret["smtputf8"]:
        ret["email_ascii"] = ret["local"] + "@" + ret["domain"]

    return ret


def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=False):
    # Validates the local part of an email address.

    if len(local) == 0:
        if not allow_empty_local:
            raise EmailSyntaxError("There must be something before the @-sign.")
        else:
            # The caller allows an empty local part. Useful for validating certain
            # Postfix aliases.
            return {
                "local": local,
                "smtputf8": False,
            }

    # RFC 5321 4.5.3.1.1
    if len(local) > 64:
        raise EmailSyntaxError("The email address is too long before the @-sign.")

    # Check the local part against the regular expression for the older ASCII requirements.
    m = re.match(DOT_ATOM_TEXT + "$", local)
    if m:
        # Return the local part unchanged and flag that SMTPUTF8 is not needed.
        return {
            "local": local,
            "smtputf8": False,
        }

    else:
        # The local part failed the ASCII check. Now try the extended internationalized requirements.
        m = re.match(DOT_ATOM_TEXT_UTF8 + "$", local)
        if not m:
            # It's not a valid internationalized address either. Report which characters were not valid.
            bad_chars = ', '.join(sorted(set(
                c for c in local if not re.match(u"[" + (ATEXT if not allow_smtputf8 else ATEXT_UTF8) + u"]", c)
            )))
            raise EmailSyntaxError("The email address contains invalid characters before the @-sign: %s." % bad_chars)

        # It would be valid if internationalized characters were allowed by the caller.
        raise EmailSyntaxError("Internationalized characters before the @-sign are not supported.")


def validate_email_domain_part(domain):
    # Empty?
    if len(domain) == 0:
        raise EmailSyntaxError("There must be something after the @-sign.")

    # Perform UTS-46 normalization, which includes casefolding, NFC normalization,
    # and converting all label separators (the period/full stop, fullwidth full stop,
    # ideographic full stop, and halfwidth ideographic full stop) to basic periods.
    # It will also raise an exception if there is an invalid character in the input,
    # such as "⒈" which is invalid because it would expand to include a period.
    try:
        domain = idna.uts46_remap(domain, std3_rules=False, transitional=False)
    except idna.IDNAError as e:
        raise EmailSyntaxError("The domain name %s contains invalid characters (%s)." % (domain, str(e)))

    # Now we can perform basic checks on the use of periods (since equivalent
    # symbols have been mapped to periods). These checks are needed because the
    # IDNA library doesn't handle well domains that have empty labels (i.e. initial
    # dot, trailing dot, or two dots in a row).
    if domain.endswith("."):
        raise EmailSyntaxError("An email address cannot end with a period.")
    if domain.startswith("."):
        raise EmailSyntaxError("An email address cannot have a period immediately after the @-sign.")
    if ".." in domain:
        raise EmailSyntaxError("An email address cannot have two periods in a row.")

    # Regardless of whether international characters are actually used,
    # first convert to IDNA ASCII. For ASCII-only domains, the transformation
    # does nothing. If internationalized characters are present, the MTA
    # must either support SMTPUTF8 or the mail client must convert the
    # domain name to IDNA before submission.
    #
    # Unfortunately this step incorrectly 'fixes' domain names with leading
    # periods by removing them, so we have to check for this above. It also gives
    # a funky error message ("No input") when there are two periods in a
    # row, also checked separately above.
    try:
        domain = idna.encode(domain, uts46=False).decode("ascii")
    except idna.IDNAError as e:
        raise EmailSyntaxError("The domain name %s contains invalid characters (%s)." % (domain, str(e)))

    # We may have been given an IDNA ASCII domain to begin with. Check
    # that the domain actually conforms to IDNA. It could look like IDNA
    # but not be actual IDNA. For ASCII-only domains, the conversion out
    # of IDNA just gives the same thing back.
    #
    # This gives us the canonical internationalized form of the domain,
    # which we should use in all error messages.
    try:
        domain_i18n = idna.decode(domain.encode('ascii'))
    except idna.IDNAError as e:
        raise EmailSyntaxError("The domain name %s is not valid IDNA (%s)." % (domain, str(e)))

    # RFC 5321 4.5.3.1.2
    if len(domain) > 255:
        raise EmailSyntaxError("The email address is too long after the @-sign.")

    # A "dot atom text", per RFC 2822 3.2.4, but using the restricted
    # characters allowed in a hostname (see ATEXT_HOSTNAME above).
    DOT_ATOM_TEXT = ATEXT_HOSTNAME + r'(?:\.' + ATEXT_HOSTNAME + r')*'

    # Check the regular expression. This is probably entirely redundant
    # with idna.decode, which also checks this format.
    m = re.match(DOT_ATOM_TEXT + "$", domain)
    if not m:
        raise EmailSyntaxError("The email address contains invalid characters after the @-sign.")

    # All publicly deliverable addresses have domain named with at least
    # one period. We also know that all TLDs end with a letter.
    if "." not in domain:
        raise EmailSyntaxError("The domain name %s is not valid. It should have a period." % domain_i18n)
    if not re.search(r"[A-Za-z]$", domain):
        raise EmailSyntaxError(
            "The domain name %s is not valid. It is not within a valid top-level domain." % domain_i18n
        )

    # Return the IDNA ASCII-encoded form of the domain, which is how it
    # would be transmitted on the wire (except when used with SMTPUTF8
    # possibly), as well as the canonical Unicode form of the domain,
    # which is better for display purposes. This should also take care
    # of RFC 6532 section 3.1's suggestion to apply Unicode NFC
    # normalization to addresses.
    return {
        "domain": domain,
        "domain_i18n": domain_i18n,
    }


def main():
    import sys
    import json

    if len(sys.argv) == 1:
        # Read lines for STDIN and validate the email address on each line.
        allow_smtputf8 = True
        for line in sys.stdin:
            try:
                email = line.strip()
                if sys.version_info < (3,):
                    email = email.decode("utf8")  # assume utf8 in input
                validate_email(email, allow_smtputf8=allow_smtputf8)
            except EmailNotValidError as e:
                print(email, e)
    else:
        # Validate the email address passed on the command line.
        email = sys.argv[1]
        allow_smtputf8 = True
        check_deliverability = True
        if sys.version_info < (3,):
            email = email.decode("utf8")  # assume utf8 in input
        try:
            result = validate_email(email, allow_smtputf8=allow_smtputf8, check_deliverability=check_deliverability)
            print(json.dumps(result, indent=2, sort_keys=True, ensure_ascii=False))
        except EmailNotValidError as e:
            if sys.version_info < (3,):
                print(unicode_class(e).encode("utf8"))
            else:
                print(e)


if __name__ == "__main__":
    main()



"""
https://github.com/icholy/durationpy/blob/master/durationpy/duration.py

>>> int(from_str('0').total_seconds())
0
>>> int(from_str('60s').total_seconds())
60
>>> int(from_str('3m30s').total_seconds())
210
>>> int(from_str('2h3m30s').total_seconds())
7410
>>> int(from_str('1d10s').total_seconds())
86410
>>> int(from_str('-1s').total_seconds())
-1
>>> print(to_str(datetime.timedelta(seconds=5)))
5s
>>> print(to_str(datetime.timedelta(seconds=59)))
59s
>>> print(to_str(datetime.timedelta(seconds=61)))
1m1s
>>> print(to_str(datetime.timedelta(seconds=3600)))
1h
>>> print(to_str(datetime.timedelta(seconds=3600 + 60 + 5)))
1h1m5s
>>> print(to_str(datetime.timedelta(seconds=23*60*60 + 60 + 5)))
23h1m5s
>>> print(to_str(datetime.timedelta(seconds=0.005)))
5ms
>>> print(to_str(datetime.timedelta(seconds=0.000005)))
5us
>>> D = 24*60*60
>>> print(to_str(datetime.timedelta(seconds=2*D + 60 + 5), extended=True))
2d1m5s
>>> print(to_str(datetime.timedelta(seconds=35*D + 60 + 5), extended=True))
1mo5d1m5s
>>> print(to_str(datetime.timedelta(seconds=370*D + 60 + 5), extended=True))
1y5d1m5s
>>> from_str('10x')
Traceback (most recent call last):
...
validr._vendor.durationpy.DurationError: Unknown unit x in duration 10x
>>> from_str('1..0h')
Traceback (most recent call last):
...
validr._vendor.durationpy.DurationError: Invalid value 1..0 in duration 1..0h
"""
import re
import datetime

_nanosecond_size = 1
_microsecond_size = 1000 * _nanosecond_size
_millisecond_size = 1000 * _microsecond_size
_second_size = 1000 * _millisecond_size
_minute_size = 60 * _second_size
_hour_size = 60 * _minute_size
_day_size = 24 * _hour_size
_week_size = 7 * _day_size
_month_size = 30 * _day_size
_year_size = 365 * _day_size

_units = {
    "ns": _nanosecond_size,
    "us": _microsecond_size,
    "µs": _microsecond_size,
    "μs": _microsecond_size,
    "ms": _millisecond_size,
    "s": _second_size,
    "m": _minute_size,
    "h": _hour_size,
    "d": _day_size,
    "w": _week_size,
    "mm": _month_size,  # deprecated, use 'mo' instead
    "mo": _month_size,
    "y": _year_size,
}


class DurationError(ValueError):
    """duration error"""


_RE_DURATION = re.compile(r'([\d\.]+)([a-zµμ]+)')


def from_str(duration: str) -> datetime.timedelta:
    """Parse a duration string to a datetime.timedelta"""

    duration = duration.strip()

    if duration in ("0", "+0", "-0"):
        return datetime.timedelta()

    total = 0
    sign = -1 if duration[0] == '-' else 1
    matches = _RE_DURATION.findall(duration)

    if not len(matches):
        raise DurationError("Invalid duration {}".format(duration))

    for (value, unit) in matches:
        if unit not in _units:
            err_msg = "Unknown unit {} in duration {}".format(unit, duration)
            raise DurationError(err_msg)
        try:
            total += float(value) * _units[unit]
        except Exception:
            err_msg = "Invalid value {} in duration {}".format(value, duration)
            raise DurationError(err_msg)

    microseconds = total / _microsecond_size
    return datetime.timedelta(microseconds=sign * microseconds)


def to_str(delta: datetime.timedelta, extended=False) -> str:
    """Format a datetime.timedelta to a duration string"""

    total_seconds = delta.total_seconds()
    sign = "-" if total_seconds < 0 else ""
    nanoseconds = abs(total_seconds * _second_size)

    if total_seconds < 1:
        result_str = _to_str_small(nanoseconds, extended)
    else:
        result_str = _to_str_large(nanoseconds, extended)

    return "{}{}".format(sign, result_str)


def _to_str_small(nanoseconds, extended):

    result_str = ""

    if not nanoseconds:
        return "0"

    milliseconds = int(nanoseconds / _millisecond_size)
    if milliseconds:
        nanoseconds -= _millisecond_size * milliseconds
        result_str += "{:g}ms".format(milliseconds)

    microseconds = int(nanoseconds / _microsecond_size)
    if microseconds:
        nanoseconds -= _microsecond_size * microseconds
        result_str += "{:g}us".format(microseconds)

    if nanoseconds:
        result_str += "{:g}ns".format(nanoseconds)

    return result_str


def _to_str_large(nanoseconds, extended):

    result_str = ""

    if extended:

        years = int(nanoseconds / _year_size)
        if years:
            nanoseconds -= _year_size * years
            result_str += "{:g}y".format(years)

        months = int(nanoseconds / _month_size)
        if months:
            nanoseconds -= _month_size * months
            result_str += "{:g}mo".format(months)

        days = int(nanoseconds / _day_size)
        if days:
            nanoseconds -= _day_size * days
            result_str += "{:g}d".format(days)

    hours = int(nanoseconds / _hour_size)
    if hours:
        nanoseconds -= _hour_size * hours
        result_str += "{:g}h".format(hours)

    minutes = int(nanoseconds / _minute_size)
    if minutes:
        nanoseconds -= _minute_size * minutes
        result_str += "{:g}m".format(minutes)

    seconds = float(nanoseconds) / float(_second_size)
    if seconds:
        nanoseconds -= _second_size * seconds
        result_str += "{:g}s".format(seconds)

    return result_str



"""A simple, fast, extensible python library for data validation."""
from .exception import Invalid, ModelInvalid, SchemaError, ValidrError, mark_index, mark_key
from .validator import (
    create_re_validator, create_enum_validator, builtin_validators, validator)
from .schema import Schema, Compiler, T, Builder
from .model import modelclass, fields, asdict, ImmutableInstanceError

__all__ = (
    'ValidrError', 'Invalid', 'ModelInvalid', 'SchemaError',
    'mark_index', 'mark_key',
    'create_re_validator', 'create_enum_validator',
    'builtin_validators', 'validator',
    'Schema', 'Compiler', 'T', 'Builder',
    'modelclass', 'fields', 'asdict', 'ImmutableInstanceError',
)



from terminaltables import AsciiTable


_NOT_SET = object()


def shorten(text, length):
    if len(text) > length:
        return text[:length] + '..'
    return text


def _format_value(value):
    if isinstance(value, str):
        return repr(shorten(value, 75))
    else:
        return shorten(str(value), 75)


def _format_error(args, position, value_clause=None):
    msg = str(args[0]) if args else 'invalid'
    if position:
        msg = '%s: %s' % (position, msg)
    if value_clause:
        msg = '%s, %s' % (msg, value_clause)
    return msg


class ValidrError(ValueError):
    """Base exception of validr"""

    def __init__(self, *args, value=_NOT_SET, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = value
        # marks item: (is_key, index_or_key)
        self.marks = []

    def mark_index(self, index=-1):
        self.marks.append((False, index))
        return self

    def mark_key(self, key):
        self.marks.append((True, key))
        return self

    @property
    def has_value(self):
        """Check has value set"""
        return self._value is not _NOT_SET

    def set_value(self, value):
        """Set value if not set"""
        if self._value is _NOT_SET:
            self._value = value

    @property
    def value(self):
        """The invalid value"""
        if self._value is _NOT_SET:
            return None
        return self._value

    @property
    def field(self):
        """First level index or key, usually it's the field"""
        if not self.marks:
            return None
        __, index_or_key = self.marks[-1]
        return index_or_key

    @property
    def position(self):
        """A string which represent the position of invalid.

        For example:

            {
                "tags": ["ok", "invalid"],  # tags[1]
                "user": {
                    "name": "invalid",      # user.name
                    "age": 500              # user.age
                }
            }
        """
        text = ''
        # cdef bint is_key
        for is_key, index_or_key in reversed(self.marks):
            if is_key:
                text = '%s.%s' % (text, index_or_key)
            else:
                if index_or_key == -1:
                    text = '%s[]' % text
                else:
                    text = '%s[%d]' % (text, index_or_key)
        if text and text[0] == '.':
            text = text[1:]
        return text

    @property
    def message(self):
        """Error message"""
        if self.args:
            return self.args[0]
        else:
            return None

    def __str__(self):
        return _format_error(self.args, self.position)


class Invalid(ValidrError):
    """Data invalid"""
    def __str__(self):
        value_clause = None
        if self.has_value:
            value_clause = 'value=%s' % _format_value(self.value)
        return _format_error(self.args, self.position, value_clause)


class ModelInvalid(Invalid):
    """Model data invalid"""
    def __init__(self, errors):
        if not errors:
            raise ValueError('errors is required')
        self.errors = errors
        message = errors[0].message or 'invalid'
        message += ' ...total {} errors'.format(len(errors))
        super().__init__(message)

    def __str__(self):
        error_items = [(ex.position, ex.message) for ex in self.errors]
        table = [("Key", "Error")] + error_items
        return '\n' + AsciiTable(table).table


class SchemaError(ValidrError):
    """Schema error"""
    def __str__(self):
        value_clause = None
        if self.has_value:
            value_clause = 'schema=%s' % self.value.repr(prefix=False, desc=False)
        return _format_error(self.args, self.position, value_clause)


class mark_index:
    """Add current index to Invalid/SchemaError"""

    # cdef int index

    def __init__(self, index=-1):
        """index = -1 means the position is uncertainty"""
        self.index = index

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None and issubclass(exc_type, ValidrError):
            exc_val.mark_index(self.index)


class mark_key:
    """Add current key to Invalid/SchemaError"""

    # cdef str key

    def __init__(self, key):
        self.key = key

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None and issubclass(exc_type, ValidrError):
            exc_val.mark_key(self.key)



import typing


class Builder:
    def __getitem__(self, keys: typing.Iterable[str]) -> "Builder": ...
    def __getattr__(self, name: str) -> "Builder": ...
    def __call__(self, *args, **kwargs) -> "Builder": ...


T: Builder



import os
from os.path import dirname, basename, splitext
from glob import glob
from setuptools import Extension, setup


def _read_file(filepath):
    with open(os.path.join(dirname(__file__), filepath), 'r', encoding='utf-8') as f:
        return f.read()


_SETUP_OPTIONS = dict(
    name='validr',
    version='1.2.1',
    keywords='validation validator validate schema jsonschema',
    description=(
        'A simple, fast, extensible python library for data validation.'),
    long_description=_read_file('README.md'),
    long_description_content_type="text/markdown",
    author='guyskk',
    author_email='guyskk@qq.com',
    url='https://github.com/guyskk/validr',
    license='MIT',
    packages=['validr', 'validr._vendor'],
    package_dir={'': 'src'},
    package_data={'validr': ['*.pyi']},
    include_package_data=True,
    install_requires=[
        'idna>=2.5',
        'pyparsing>=2.1.0',
        'terminaltables>=3.1.0',
    ],
    extras_require={
        'dev': [
            'pre-commit>=0.13.3',
            'flake8>=3.2.1',
            'pytest>=3.0.6',
            'pytest-cov>=2.4.0',
            'codecov>=2.0.5',
            'invoke>=1.0.0',
            'twine>=1.11.0',
        ],
        'benchmark': [
            'click>=6.7',
            'schema>=0.6.5',
            'jsonschema>=2.5.1',
            'schematics>=2.0.0a1',
            'voluptuous>=0.9.3',
        ],
    },
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)


_SETUP_MODES = {
    'pyx',       # cythonize *.pyx
    'pyx_dbg',   # cythonize *.pyx with debug info
    'c',         # ext_modules from *.c
    'c_dbg',     # ext_modules from *.c with debug info
    'py',        # pure python
    'dist',      # build *_c.c and *_py.py for release
    'dist_dbg',  # build *_c.c and *_py.py for release with debug info
}


def _has_c_compiler():
    try:
        import distutils.ccompiler
        cc = distutils.ccompiler.new_compiler()
        return cc.has_function('rand', includes=['stdlib.h'])
    except Exception as ex:
        print('failed to check c compiler: {}'.format(ex))
        return False


def _get_validr_setup_mode():
    mode = os.getenv('VALIDR_SETUP_MODE')
    if mode:
        mode = mode.strip().lower()
        assert mode in _SETUP_MODES, 'unknown validr setup mode {}'.format(mode)
        return mode
    if _has_c_compiler():
        return 'c'
    else:
        return 'py'


def _prepare_setup_options(mode):
    is_pyx = mode in ['pyx', 'pyx_dbg']
    is_c = mode in ['c', 'c_dbg']
    is_dist = mode in ['dist', 'dist_dbg']
    is_debug = mode.endswith('_dbg')
    ext_modules = None
    if is_pyx or is_c or is_dist:
        if is_pyx or is_dist:
            from multiprocessing import cpu_count
            from Cython.Build import cythonize
            directives = {'language_level': 3}
            if is_debug:
                directives.update({
                    'profile': True,
                    'linetrace': True,
                })
            ext_modules = cythonize(
                'src/validr/*.pyx',
                nthreads=cpu_count(),
                compiler_directives=directives
            )
        if is_c:
            sources = list(glob('src/validr/*.c'))
            assert sources, 'Not found any *.c source files'
            ext_modules = []
            for filepath in sources:
                module_name = 'validr.' + splitext(basename(filepath))[0]
                ext_modules.append(Extension(module_name, [filepath]))
        if is_debug:
            for m in ext_modules:
                m.define_macros.extend([
                    ('CYTHON_TRACE', '1'),
                    ('CYTHON_TRACE_NOGIL', '1'),
                    ('CYTHON_PROFILE', '1'),
                ])
    if is_dist:
        from validr_uncython import compile_pyx_to_py
        sources = list(glob('src/validr/*.pyx'))
        compile_pyx_to_py(sources, debug=is_debug)

    return dict(ext_modules=ext_modules, **_SETUP_OPTIONS)


def _validr_setup():
    mode = _get_validr_setup_mode()
    print('VALIDR_SETUP_MODE={}'.format(mode))
    options = _prepare_setup_options(mode)
    setup(**options)


_validr_setup()



"""
Model class is a convenient way to use schema, it's inspired by data class but
works differently.
"""
from .schema import Compiler, T, Schema
from .exception import Invalid, ModelInvalid, mark_key
from .validator import is_dict, get_dict_value, get_object_value


class ImmutableInstanceError(AttributeError):
    """Raised when an attempt is modify a immutable class"""


def modelclass(cls=None, *, compiler=None, immutable=False):
    if cls is not None:
        return _create_model_class(cls, compiler, immutable)

    def decorator(cls):
        return _create_model_class(cls, compiler, immutable)

    return decorator


def _value_asdict(value):
    if hasattr(value, '__asdict__'):
        return value.__asdict__()
    elif is_dict(value):
        return {k: _value_asdict(v) for k, v in value.items()}
    elif isinstance(value, (list, tuple, set)):
        return [_value_asdict(x) for x in value]
    else:
        return value


def _extract_schemas(cls):
    schemas = {}
    for k, v in vars(cls).items():
        if k == "__schema__":
            continue
        if hasattr(v, "__schema__"):
            v = v.__schema__
        if isinstance(v, Schema):
            schemas[k] = v
    return schemas


def _extract_post_init(cls):
    f = vars(cls).get("__post_init__", None)
    if f is None or not callable(f):
        return None
    return f


class Field:
    def __init__(self, name, schema, compiler):
        self.name = name
        self.__schema__ = schema
        with mark_key(self.name):
            self.validate = compiler.compile(schema)

    def __repr__(self):
        info = "schema={!r}".format(self.__schema__)
        return "Field(name={!r}, {})".format(self.name, info)

    def __get__(self, obj, obj_type):
        if obj is None:
            return self
        return obj.__dict__.get(self.name, None)

    def __set__(self, obj, value):
        with mark_key(self.name):
            value = self.validate(value)
        obj.__dict__[self.name] = value


def _create_model_class(model_cls, compiler, immutable):

    compiler = compiler or Compiler()

    class ModelMeta(type):
        def __init__(cls, *args, **kwargs):
            super().__init__(*args, **kwargs)
            schemas = {}
            post_inits = []
            for cls_or_base in reversed(cls.__mro__):
                post_init = _extract_post_init(cls_or_base)
                if post_init is not None:
                    post_inits.append(post_init)
                for name, schema in _extract_schemas(cls_or_base).items():
                    schemas[name] = schema
            for name, schema in schemas.items():
                setattr(cls, name, Field(name, schema, compiler))
            cls.__post_inits = post_inits
            cls.__schema__ = T.dict(schemas).__schema__
            cls.__fields__ = frozenset(schemas)

        def post_init(cls, instance):
            for post_init in cls.__post_inits:
                post_init(instance)

        def __repr__(cls):
            # use __schema__ can keep fields order in python>=3.6
            fields = ", ".join(cls.__schema__.items)
            return "{}<{}>".format(cls.__name__, fields)

        def __getitem__(self, keys):
            s = self.__schema__
            items = s.items or {}
            if not isinstance(keys, (list, tuple, set, frozenset)):
                if keys not in items:
                    raise KeyError("key {!r} is not exists".format(keys))
                return items[keys]
            schema = Schema(validator=s.validator, params=s.params.copy())
            schema.items = {}
            for k in keys:
                if k not in items:
                    raise KeyError("key {!r} is not exists".format(k))
                schema.items[k] = items[k]
            return T(schema)

    class Model(model_cls, metaclass=ModelMeta):

        if "__init__" not in model_cls.__dict__:

            def __init__(self, *obj, **params):
                self.__dict__["__immutable__"] = False
                params_set = set(params)
                errors = []
                if obj:
                    if len(obj) > 1:
                        msg = (
                            "__init__() takes 2 positional arguments "
                            "but {} were given".format(len(obj) + 1)
                        )
                        raise TypeError(msg)
                    obj = obj[0]
                    if is_dict(obj):
                        getter = get_dict_value
                    else:
                        getter = get_object_value
                    for k in self.__fields__ - params_set:
                        try:
                            setattr(self, k, getter(obj, k))
                        except Invalid as ex:
                            errors.append(ex)
                else:
                    for k in self.__fields__ - params_set:
                        try:
                            setattr(self, k, None)
                        except Invalid as ex:
                            errors.append(ex)
                for k in self.__fields__ & params_set:
                    try:
                        setattr(self, k, params[k])
                    except Invalid as ex:
                        errors.append(ex)
                for k in params_set - self.__fields__:
                    errors.append(Invalid("undesired key").mark_key(k))
                if errors:
                    raise ModelInvalid(errors)
                type(self).post_init(self)
                self.__dict__["__immutable__"] = immutable

        else:

            def __init__(self, *args, **kwargs):
                self.__dict__["__immutable__"] = False
                super().__init__(*args, **kwargs)
                type(self).post_init(self)
                self.__dict__["__immutable__"] = immutable

        if immutable:

            def __setattr__(self, name, value):
                if self.__immutable__:
                    msg = "{} object is immutable!".format(type(self).__name__)
                    raise ImmutableInstanceError(msg)
                return object.__setattr__(self, name, value)

            def __delattr__(self, name):
                if self.__immutable__:
                    msg = "{} object is immutable!".format(type(self).__name__)
                    raise ImmutableInstanceError(msg)
                return object.__delattr__(self, name)

        if "__repr__" not in model_cls.__dict__:

            def __repr__(self):
                params = []
                # use __schema__ can keep fields order
                for k in self.__schema__.items:
                    v = getattr(self, k)
                    params.append("{}={!r}".format(k, v))
                params = ", ".join(params)
                return "{}({})".format(type(self).__name__, params)

        if "__eq__" not in model_cls.__dict__:

            def __eq__(self, other):
                fields = getattr(other, "__fields__", None)
                if not fields:
                    return False
                if self.__fields__ != fields:
                    return False
                for k in self.__fields__:
                    if getattr(self, k, None) != getattr(other, k, None):
                        return False
                return True

        def __asdict__(self, *, keys=None):
            if not keys:
                keys = self.__fields__
            else:
                keys = set(keys) & self.__fields__
            ret = {}
            for k in keys:
                v = getattr(self, k)
                if v is not None:
                    v = _value_asdict(v)
                ret[k] = v
            return ret

    Model.__module__ = model_cls.__module__
    Model.__name__ = model_cls.__name__
    Model.__qualname__ = model_cls.__qualname__
    Model.__doc__ = model_cls.__doc__

    return Model


def fields(m) -> set:
    """Get fields of model or dict schema"""
    if hasattr(m, '__fields__'):  # modelclass
        return m.__fields__
    if hasattr(m, '__schema__'):
        schema = m.__schema__     # T.dict({...})
    else:
        schema = m                # Schema
    if isinstance(schema, Schema):
        if schema.validator == 'dict':
            if schema.items:
                return set(schema.items.keys())
            else:
                return set()
    raise TypeError("can not find fields of {!r}".format(m))


def asdict(m, *, keys=None) -> dict:
    """Convert model instance to dict"""
    return m.__asdict__(keys=keys)



"""
Schema and Compiler

schema is instance of Schema or object which has __schema__ attribute,
and the __schema__ is instance of Schema.

compiler can compile schema:

    compiler.compile(schema) -> validate function.

validate function has __schema__ attribute, it's also a schema.

the Builder's instance T can build schema.
in addition, T can be called directly, which convert schema-like things to
instance of Builder:

    T(JSON) -> Isomorph Schema
    T(func) -> Schema of validate func
    T(Schema) -> Copy of Schema
    T(Model) -> Schema of Model

Builder support schema slice:

    T[...keys] -> sub schema

relations:

    T(schema) -> T
    T.__schema__ -> Schema
"""
import json
import copy
import enum
import inspect

from pyparsing import (
    Group, Keyword, Optional, StringEnd, StringStart, Suppress,
    ZeroOrMore, quotedString, removeQuotes, replaceWith,
    pyparsing_common, ParseBaseException,
)

from .validator import builtin_validators
from .exception import SchemaError, mark_index, mark_key


def _make_keyword(kwd_str, kwd_value):
    return Keyword(kwd_str).setParseAction(replaceWith(kwd_value))


def _define_value():
    TRUE = _make_keyword('true', True)
    FALSE = _make_keyword('false', False)
    NULL = _make_keyword('null', None)
    STRING = quotedString().setParseAction(removeQuotes)
    NUMBER = pyparsing_common.number()
    return TRUE | FALSE | NULL | STRING | NUMBER


def _define_element():
    VALIDATOR = pyparsing_common.identifier.setName('validator').setResultsName('validator')
    ITEMS = _define_value().setName('items').setResultsName('items')
    ITEMS_WRAPPER = Optional(Suppress('(') + ITEMS + Suppress(')'))
    PARAMS_KEY = pyparsing_common.identifier.setName('key').setResultsName('key')
    PARAMS_VALUE = _define_value().setName('value').setResultsName('value')
    PARAMS_VALUE_WRAPPER = Optional(Suppress('(') + PARAMS_VALUE + Suppress(')'))
    PARAMS_KEY_VALUE = Group(Suppress('.') + PARAMS_KEY + PARAMS_VALUE_WRAPPER)
    PARAMS = Group(ZeroOrMore(PARAMS_KEY_VALUE)).setName('params').setResultsName('params')
    return StringStart() + VALIDATOR + ITEMS_WRAPPER + PARAMS + StringEnd()


ELEMENT_GRAMMAR = _define_element()


def _dump_value(value):
    if value is None:
        return 'null'
    elif value is False:
        return 'false'
    elif value is True:
        return 'true'
    elif isinstance(value, str):
        return repr(value)  # single quotes by default
    elif isinstance(value, Schema):
        return value.validator
    else:
        return str(value)  # number


def _pair(k, v):
    return '{}({})'.format(k, _dump_value(v))


def _sort_schema_params(params):
    def key(item):
        k, v = item
        if k == 'desc':
            return 3
        if k == 'optional':
            return 2
        if k == 'default':
            return 1
        if isinstance(v, bool):
            return -1
        if isinstance(v, str):
            return -2
        else:
            return -3
    return list(sorted(params, key=key))


def _schema_of(obj) -> "Schema":
    if hasattr(obj, '__schema__'):
        obj = obj.__schema__
    return obj


def _schema_copy_of(obj) -> "Schema":
    if isinstance(obj, Schema):
        obj = obj.copy()
    return obj


def _schema_primitive_of(obj):
    if isinstance(obj, Schema):
        obj = obj.to_primitive()
    return obj


def _is_model(obj):
    return inspect.isclass(obj) and hasattr(obj, '__schema__')


class Schema:

    def __init__(self, *, validator=None, items=None, params=None):
        self.validator = validator
        self.items = items
        self.params = params or {}

    def __eq__(self, other):
        other = _schema_of(other)
        if not isinstance(other, Schema):
            return False
        return (self.validator == other.validator and
                self.items == other.items and
                self.params == other.params)

    def __hash__(self):
        params = tuple(sorted(self.params.items()))
        items = self.items
        if isinstance(items, dict):
            items = tuple(sorted(items.items()))
        elif isinstance(items, list):
            items = tuple(items)
        return hash((self.validator, items, params))

    def __str__(self):
        return json.dumps(self.to_primitive(), indent=4,
                          ensure_ascii=False, sort_keys=True)

    def repr(self, *, prefix=True, desc=True):
        if not self.validator:
            return 'T' if prefix else ''
        ret = ['T'] if prefix else []
        if self.items is None:
            ret.append(self.validator)
        else:
            if self.validator == 'dict':
                keys = ', '.join(sorted(self.items)) if self.items else ''
                ret.append('{}({})'.format(self.validator, '{' + keys + '}'))
            elif self.validator == 'list':
                ret.append('{}({})'.format(self.validator, self.items.validator))
            elif self.validator == 'enum':
                values = ', '.join(map(_dump_value, self.items)) if self.items else ''
                ret.append('{}({})'.format(self.validator, '{' + values + '}'))
            elif self.validator == 'union':
                if self.items and isinstance(self.items, list):
                    keys = ', '.join(x.validator for x in self.items)
                    ret.append('{}([{}])'.format(self.validator, keys))
                else:
                    keys = ', '.join(sorted(self.items)) if self.items else ''
                    ret.append('{}({})'.format(self.validator, '{' + keys + '}'))
            elif self.validator == 'model' and self.items is not None:
                ret.append('{}({})'.format(self.validator, self.items.__name__))
            else:
                ret.append(_pair(self.validator, self.items))
        for k, v in _sort_schema_params(self.params.items()):
            if not desc and k == 'desc':
                continue
            if v is False:
                continue
            if v is True:
                ret.append(k)
            else:
                ret.append(_pair(k, v))
        return '.'.join(ret)

    def __repr__(self):
        r = self.repr(prefix=False)
        return '{}<{}>'.format(type(self).__name__, r)

    def copy(self):
        params = {k: _schema_copy_of(v)for k, v in self.params.items()}
        schema = type(self)(validator=self.validator, params=params)
        if self.validator == 'dict' and self.items is not None:
            items = {k: _schema_copy_of(v) for k, v in self.items.items()}
        elif self.validator == 'list' and self.items is not None:
            items = _schema_copy_of(self.items)
        elif self.validator == 'union' and self.items is not None:
            if isinstance(self.items, list):
                items = [_schema_copy_of(x) for x in self.items]
            else:
                items = {k: _schema_copy_of(v) for k, v in self.items.items()}
        else:
            items = copy.copy(self.items)
        schema.items = items
        return schema

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def to_primitive(self):
        if not self.validator:
            return None
        # T.model not support in JSON, convert it to T.dict
        if self.validator == 'model':
            if self.items is None:
                items = None
            else:
                items = _schema_of(self.items).items
            self = Schema(validator='dict', items=items, params=self.params)
        ret = []
        if self.validator in {'dict', 'list', 'union', 'enum'} or self.items is None:
            ret.append(self.validator)
        else:
            ret.append(_pair(self.validator, self.items))
        for k, v in _sort_schema_params(self.params.items()):
            if self.validator == 'dict' and k in {'key', 'value'}:
                continue
            if v is False:
                continue
            if v is True:
                ret.append(k)
            else:
                ret.append(_pair(k, v))
        ret = '.'.join(ret)
        if self.validator == 'dict':
            ret = {'$self': ret}
            for pkey in ['key', 'value']:
                pvalue = self.params.get(pkey)
                if pvalue is not None:
                    ret['$self_{}'.format(pkey)] = _schema_primitive_of(pvalue)
            if self.items is not None:
                for k, v in self.items.items():
                    ret[k] = _schema_primitive_of(v)
        elif self.validator == 'list' and self.items is not None:
            ret = [ret, _schema_primitive_of(self.items)]
        elif self.validator == 'enum' and self.items is not None:
            ret = [ret, *self.items]
        elif self.validator == 'union' and self.items is not None:
            if isinstance(self.items, list):
                ret = [ret]
                for x in self.items:
                    ret.append(_schema_primitive_of(x))
            else:
                ret = {'$self': ret}
                for k, v in self.items.items():
                    ret[k] = _schema_primitive_of(v)
        return ret

    @classmethod
    def parse_element(cls, text):
        if text is None:
            raise SchemaError("can't parse None")
        text = text.strip()
        if not text:
            raise SchemaError("can't parse empty string")
        try:
            result = ELEMENT_GRAMMAR.parseString(text, parseAll=True)
        except ParseBaseException as ex:
            msg = 'invalid syntax in col {} of {!r}'.format(ex.col, repr(ex.line))
            raise SchemaError(msg) from None
        validator = result['validator']
        items = None
        if 'items' in result:
            items = result['items']
        params = {}
        for item in result['params']:
            value = True
            if 'value' in item:
                value = item['value']
            params[item['key']] = value
        return cls(validator=validator, items=items, params=params)

    @classmethod
    def parse_isomorph_schema(cls, obj):
        if isinstance(obj, str):
            return cls.parse_element(obj)
        elif isinstance(obj, dict):
            e = cls.parse_element(obj.pop('$self', 'dict'))
            items = {}
            for k, v in obj.items():
                with mark_key(k):
                    items[k] = cls.parse_isomorph_schema(v)
            for pkey in ['key', 'value']:
                pvalue = items.pop('$self_{}'.format(pkey), None)
                if pvalue is not None:
                    e.params[pkey] = pvalue
            return cls(validator=e.validator, items=items, params=e.params)
        elif isinstance(obj, list):
            if len(obj) == 1:
                validator = 'list'
                params = None
                items = cls.parse_isomorph_schema(obj[0])
            elif len(obj) >= 2:
                e = cls.parse_element(obj[0])
                validator = e.validator
                params = e.params
                if validator == 'list':
                    if len(obj) > 2:
                        raise SchemaError('invalid list schema')
                    with mark_index():
                        items = cls.parse_isomorph_schema(obj[1])
                elif validator == 'enum':
                    items = list(obj[1:])
                elif validator == 'union':
                    items = []
                    for i, x in enumerate(obj[1:]):
                        with mark_index(i):
                            items.append(cls.parse_isomorph_schema(x))
                else:
                    raise SchemaError('unknown {} schema'.format(validator))
            else:
                raise SchemaError('invalid list schema')
            return cls(validator=validator, items=items, params=params)
        else:
            raise SchemaError('{} object is not schema'.format(type(obj)))


class Compiler:

    def __init__(self, validators=None, is_dump=False):
        self.validators = builtin_validators.copy()
        if validators:
            self.validators.update(validators)
        self.is_dump = is_dump

    def compile(self, schema):
        schema = _schema_of(schema)
        if not isinstance(schema, Schema):
            raise SchemaError('{} object is not schema'.format(type(schema)))
        if not schema.validator:
            raise SchemaError('incomplete schema')
        validator = self.validators.get(schema.validator)
        if not validator:
            raise SchemaError('validator {!r} not found'.format(schema.validator))
        return validator(self, schema)


_BUILDER_INIT = 'init'
_EXP_ATTR = 'expect-attr'
_EXP_ATTR_OR_ITEMS = 'expect-attr-or-items'
_EXP_ATTR_OR_CALL = 'expect-attr-or-call'


class Builder:

    def __init__(self, state=_BUILDER_INIT, *, validator=None,
                 items=None, params=None, last_attr=None):
        self._state = state
        self._schema = Schema(validator=validator, items=items, params=params)
        self._last_attr = last_attr

    @property
    def __schema__(self):
        return self._schema

    def __repr__(self):
        return self._schema.repr()

    def __str__(self):
        return self._schema.__str__()

    def __eq__(self, other):
        return self._schema == _schema_of(other)

    def __hash__(self):
        return self._schema.__hash__()

    def __getitem__(self, keys):
        if not self._schema.validator:
            raise ValueError('can not slice empty schema')
        if self._schema.validator != 'dict':
            raise ValueError('can not slice non-dict schema')
        if not isinstance(keys, (list, tuple)):
            keys = (keys,)
        schema = Schema(validator=self._schema.validator,
                        params=self._schema.params.copy())
        schema.items = {}
        items = self._schema.items or {}
        for k in keys:
            if k not in items:
                raise ValueError('key {!r} is not exists'.format(k))
            schema.items[k] = items[k]
        return T(schema)

    def __getattr__(self, name):
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError('{!r} object has no attribute {!r}'.format(
                type(self).__name__, name))
        if self._state == _BUILDER_INIT:
            return Builder(_EXP_ATTR_OR_ITEMS, validator=name)
        else:
            params = self._schema.params.copy()
            params[name] = True
            return Builder(
                _EXP_ATTR_OR_CALL, validator=self._schema.validator,
                items=self._schema.items, params=params, last_attr=name)

    def __call__(self, *args, **kwargs):
        if self._state == _BUILDER_INIT:
            return self._load_schema(*args, **kwargs)
        if self._state not in [_EXP_ATTR_OR_ITEMS, _EXP_ATTR_OR_CALL]:
            raise SchemaError('current state not callable')
        if self._state == _EXP_ATTR_OR_ITEMS:
            if args and kwargs:
                raise SchemaError("can't call with both positional argument and keyword argument")
            if len(args) > 1:
                raise SchemaError("can't call with more than one positional argument")
            if self._schema.validator in {'dict', 'union'}:
                items = args[0] if args else kwargs
            elif self._schema.validator == 'model':
                if len(args) != 1 or kwargs:
                    raise SchemaError('require exactly one positional argument')
                items = args[0]
            else:
                if kwargs:
                    raise SchemaError("can't call with keyword argument")
                if not args:
                    raise SchemaError('require one positional argument')
                items = args[0]
            items = self._check_items(items)
            params = self._schema.params
        else:
            if kwargs:
                raise SchemaError("can't call with keyword argument")
            if not args:
                raise SchemaError('require one positional argument')
            if len(args) > 1:
                raise SchemaError("can't call with more than one positional argument")
            param_value = self._check_param_value(self._last_attr, args[0])
            items = self._schema.items
            params = self._schema.params.copy()
            params[self._last_attr] = param_value
        return Builder(_EXP_ATTR, validator=self._schema.validator,
                       items=items, params=params, last_attr=None)

    def _load_schema(self, obj):
        obj = _schema_of(obj)
        if isinstance(obj, Schema):
            obj = obj.copy()
        elif isinstance(obj, (str, list, dict)):
            obj = Schema.parse_isomorph_schema(obj)
        else:
            raise SchemaError('{} object is not schema'.format(type(obj)))
        if not obj.validator:
            state = _BUILDER_INIT
        elif not obj.items and not obj.params:
            state = _EXP_ATTR_OR_ITEMS
        else:
            state = _EXP_ATTR
        return Builder(state, validator=obj.validator,
                       items=obj.items, params=obj.params, last_attr=None)

    def _check_dict_items(self, items):
        if not isinstance(items, dict):
            raise SchemaError('items must be dict')
        ret = {}
        for k, v in items.items():
            v = _schema_of(v)
            if not isinstance(v, Schema):
                raise SchemaError('items[{}] is not schema'.format(k))
            ret[k] = v
        return ret

    def _check_items(self, items):
        if self._schema.validator == 'dict':
            ret = self._check_dict_items(items)
        elif self._schema.validator == 'list':
            items = _schema_of(items)
            if not isinstance(items, Schema):
                raise SchemaError('items is not schema')
            ret = items
        elif self._schema.validator == 'enum':
            if isinstance(items, str):
                items = set(items.replace(',', ' ').strip().split())
            if inspect.isclass(items) and issubclass(items, enum.Enum):
                items = [x.value for x in items.__members__.values()]
            if not isinstance(items, (list, tuple, set)):
                raise SchemaError('items is not list or set')
            ret = []
            for i, v in enumerate(items):
                if not isinstance(v, (bool, int, float, str)):
                    raise SchemaError('enum value must be bool, int, float or str')
                ret.append(v)
            ret = list(sorted(set(ret), key=lambda x: (str(type(x)), str(x))))
        elif self._schema.validator == 'union':
            if isinstance(items, list):
                ret = []
                for i, v in enumerate(items):
                    v = _schema_of(v)
                    if not isinstance(v, Schema):
                        raise SchemaError('items[{}] is not schema'.format(i))
                    ret.append(v)
            else:
                ret = self._check_dict_items(items)
        elif self._schema.validator == 'model':
            if not _is_model(items):
                raise SchemaError('items must be model class')
            ret = items
        else:
            if not isinstance(items, (bool, int, float, str)):
                raise SchemaError('items must be bool, int, float or str')
            ret = items
        return ret

    def _check_param_value(self, key, value):
        if self._schema.validator == 'dict':
            if key in {'key', 'value'}:
                return self._check_dict_param_value(key, value)
        if value is not None and not isinstance(value, (bool, int, float, str)):
            raise SchemaError('parameter value must be bool, int, float or str')
        return value

    def _check_dict_param_value(self, key, value):
        value = _schema_of(value)
        if value is not None and not isinstance(value, Schema):
            raise SchemaError('dict {} parameter is not schema'.format(key))
        return value


T = Builder()
